from pandac.PandaModules import AuxBitplaneAttrib, Vec4

from Filter import Filter, TextureInfo


class BloomFilter(Filter):
    """A bloom filter, causing bright objects to have a glowing halo around them.

    Note: If you want to use glow maps to indicate which parts of the image should receive bloom,
    you should assign a nonzero value to the alpha value of the "blend" parameter, and you should
    enable the shader generator for models that have glow maps applied. (The autoshader then writes
    the glow map into the alpha channel of the color texture; the alpha channel is otherwise unused.)

    """

    SIZES = ["small", "medium", "large"]

    def __init__(self, **kwargs):
        super(BloomFilter, self).__init__(**kwargs)

        # super's init will call cleanup(), which will call detach(), which triggers setdown(),
        # so we don't need to set self._downsampler here.

    def reset(self):
        super(BloomFilter, self).reset()
        self.sort = 0  # first within the pipeline stage (right after init, before lens flare)

        # The input texture for the bloom filter must see all previous changes;
        # hence this is not mergeable, and must come at an appropriate point
        # in the simulated image-forming process.
        #
        self.stageName   = "LensOpticsLate"
        self.isMergeable = False

        self.blend=(0.3,0.4,0.3,0.0)
        self.mintrigger=0.6
        self.maxtrigger=1.0
        self.desat=0.6
        self.intensity=1.0
        self.bloomSize="medium"

    def setup(self, phase):
        if phase == "pipeline":
            self.registerInputTexture(texName="bloom3")
            self.requireAuxBits(bitmask=AuxBitplaneAttrib.ABOGlow)

        elif phase == "stage":
            if self.bloomSize not in BloomFilter.SIZES:
                raise ValueError("Bloom size '%s' unknown; valid sizes: %s" % (self.bloomSize, BloomFilter.SIZES) )

            self.createInternalTextures( *["bloom%d" % i for i in xrange(4)] )

            if self.bloomSize == "large":
                scale = 8
                self._downsampler = "filter-down4.sha"
            elif self.bloomSize == "medium":
                scale = 4
                self._downsampler = "filter-copy.sha"
            else: # self.bloomSize == "small":
                scale = 2
                self._downsampler = "filter-copy.sha"

            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["bloom0"], div=2, align=scale))
            self.interQuads[0].setShaderInput("src", self.getTextureInfo("color").texture)

            # In setShaderInput(), we can use self.textures["bloom0"] or self.getTexture("bloom0").texture.
            # They refer to the same texture, because we have created an internal texture "bloom0".
            #
            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["bloom1"], div=scale, align=scale))
            self.interQuads[1].setShaderInput("src", self.getTextureInfo("bloom0").texture)

            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["bloom2"], div=scale, align=scale))
            self.interQuads[2].setShaderInput("src", self.getTextureInfo("bloom1").texture)

            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["bloom3"], div=scale, align=scale))
            self.interQuads[3].setShaderInput("src", self.getTextureInfo("bloom2").texture)

    def setdown(self, phase):
        # Opposite of setup(). Tear down any extra stuff set up there.
        #
        # self.interQuads and self.textures are considered "usual stuff",
        # and are cleaned up by detach() automatically.
        #
        self._downsampler = None

    def make(self):
        # Compile/load shaders of internal stages.
        self.interQuads[0].setShader(self.loadShader("filter-bloomi.sha"))
        self.interQuads[1].setShader(self.loadShader(self._downsampler))
        self.interQuads[2].setShader(self.loadShader("filter-bloomx.sha"))
        self.interQuads[3].setShader(self.loadShader("filter-bloomy.sha"))

    def synthesize(self):
        txbloom3 = self.getTextureInfo("bloom3")
        code  = "pixcolor = saturate(pixcolor);\n"
        code += "float4 bloom = 0.5 * tex2D(%(k_txbloom3)s, %(texcoord_txbloom3)s);\n" % { "k_txbloom3" : txbloom3.varname,
                                                                                           "texcoord_txbloom3" : txbloom3.texcoord }
        code += "pixcolor = 1.0 - ((1.0 - bloom)*(1.0 - pixcolor));\n"
        return ("bloomFilter", code, "// Bloom blend pass\n")


    # The internal shader parameters "blend", "trigger" and "intensity" do not map directly to the user-given
    # parameter values, but need additional calculations to produce the values actually used in the shader.
    #
    # This is implemented in the _compute*() private methods.
    #
    def _computeBlendInput(self):
        value = self._blend
        return Vec4(value[0], value[1], value[2], 2.0*value[3])

    def _computeTriggerInput(self):
        if self._maxtrigger <= self._mintrigger:
            raise ValueError("maxtrigger must be > mintrigger; currently: mintrigger = %g, maxtrigger = %g" % (self._mintrigger, self._maxtrigger))
        return Vec4(self._mintrigger, 1.0/(self._maxtrigger - self._mintrigger), 0.0, 0.0)

    def _computeIntensityInput(self):
        value = 3.0 * self._intensity
        return Vec4(value, value, value, value)


    @property
    def blend(self):
        """The bloom filter needs to measure the brightness of each pixel. It does this by
        weighting the R,G,B, and A components. Default weights: (0.3,0.4,0.3,0.0).

        You should assign a nonzero weight to the alpha channel if you want the glow map
        to have an effect, or a value like (0, 0, 0, 1) if you only want your glow map
        to indicate which parts of a model should glow.

        """
        return self._blend
    @blend.setter
    def blend(self, value):
        self._blend = value
        if len(self.interQuads)  and  self.interQuads[0] is not None:
            self.interQuads[0].setShaderInput( "blend", self._computeBlendInput() )

    @property
    def mintrigger(self):
        """Minimum brightness at which a halo is generated."""
        return self._mintrigger
    @mintrigger.setter
    def mintrigger(self, value):
        self._mintrigger = value
        if len(self.interQuads)  and  self.interQuads[0] is not None:
            self.interQuads[0].setShaderInput( "trigger", self._computeTriggerInput() )

    @property
    def maxtrigger(self):
        """Maximum brightness at which the halo reaches peak intensity."""
        return self._maxtrigger
    @maxtrigger.setter
    def maxtrigger(self, value):
        self._maxtrigger = value
        if len(self.interQuads)  and  self.interQuads[0] is not None:
            self.interQuads[0].setShaderInput( "trigger", self._computeTriggerInput() )

    @property
    def desat(self):
        """Degree to which the halo is desaturated. Float, in [0, 1].

        Setting this to zero means the halo is the same color as the bright pixel.
        Setting this to one means the halo is white.

        """
        return self._desat
    @desat.setter
    def desat(self, value):
        self._desat = value
        if len(self.interQuads)  and  self.interQuads[0] is not None:
            self.interQuads[0].setShaderInput( "desat", value )

    @property
    def intensity(self):
        """An adjustment parameter for the brightness of the halos."""
        return self._intensity
    @intensity.setter
    def intensity(self, value):
        self._intensity = value
        # This parameter goes to the last internal stage.
        if len(self.interQuads)  and  self.interQuads[3] is not None:
            self.interQuads[3].setShaderInput( "intensity", self._computeIntensityInput() )

    @property
    def bloomSize(self):
        """Halo size. One of BloomFilter.SIZES ("small", "medium" or "large").

        Controls the distance how far the halo effect reaches out from each blooming pixel.

        This is a discrete value instead of a continuous one, because the blur operation used
        internally in the bloom effect involves downsampling the original image by a power of two.

        """
        return self._bloomSize
    @bloomSize.setter
    def bloomSize(self, value):
        # Changing this parameter may change texture scalings for our internal textures,
        # so we recompile the whole pipeline.
        if (not hasattr(self, "_bloomSize")  or  value != self._bloomSize)  and  self.pipeline is not None:
            self.pipeline._needsCompile = True
        self._bloomSize = value

