from Filter import Filter, TextureInfo

# A filter with internal stages.
#
class BlurSharpenFilter(Filter):
    """A blur/sharpen filter. If the 'amount' parameter is 1.0, it will not have effect.
    A value of 0.0 means fully blurred, and a value higher than 1.0 sharpens the image."""

    def __init__(self, **kwargs):
        super(BlurSharpenFilter, self).__init__(**kwargs)

    def reset(self):
        super(BlurSharpenFilter, self).reset()
        self.sort = 0  # first within the pipeline stage (right after init)
                       # (the blur filter inherently cannot respect previous modifications to pixcolor)

        # The input texture for the blur filter must see all previous changes;
        # hence this is not mergeable, and must come at an appropriate point
        # in the simulated image-forming process.
        #
        self.stageName   = "LensFocus"
        self.isMergeable = False

        self.amount = 0.0  # full strength

    def setup(self, phase):
        if phase == "pipeline":
            # Our compositing fshader wants to blend between the original color texture provided
            # by FilterStage, and the internal blur1 texture set up here during FilterStage attach
            # (i.e. the output texture of the last internal stage).
            #
            # Register the "blur1" texture as an input texture to the compositing fshader
            # (in the FilterStage where this filter happens to be assigned), so that the code
            # provided in synthesize() will have access to it.
            #
            # We do not need to register the "color" texture, as our code does not access it explicitly,
            # and does not need its texture coordinate, either. From the color texture, we only need the
            # color of the current pixel, which is provided in pixcolor.
            #
            self.registerInputTexture(texName="blur1")

            # Our compositing fshader also has a custom input to control the strength of the blur effect.
            #
            # Note that registration only requests FilterStage to write the code to pass in this input;
            # we still need to create a run-time parameter and connect it to shader input "blur_amount"
            # (stripping the "k_" prefix).
            #
            self.registerCustomInput(inputType="float", inputName="k_blur_amount")

        elif phase == "stage":
            # Create interQuads for internal stages.
            #
            # Any setup-time parameters for the internal stages are applied here.
            # Usually this means configuring input and output textures to set up an internal pipeline.
            #
            # Note that the output texture for each interQuad is chosen in the call to renderQuadInto().
            # It must be an internal texture; only the compositing fshader can render to the final output.

            # Create output textures for internal stages.
            #
            self.createInternalTextures( "blur0", "blur1" )

            # Create a quad that outputs to texture "blur0". Connect its shader input named "src"
            # to the input color texture provided to us by the FilterStage this filter is attached to.
            #
            # It doesn't matter that the shader hasn't been created yet. Shader inputs are NodePath attributes,
            # so when we assign a shader later, the input named "src" will magically become available.
            # (It is then up to the shader code to actually use the input having that name.)
            #
            # Note that shader inputs for internal shaders are not registered to FilterStage;
            # the registration mechanism only controls code generation of the compositing fshader.)
            #
            # We use div=2 to make a quarter-resolution quad relative to the resolution of the input texture.
            # This is good enough for blur, and usually renders much faster due to much fewer fshader calls.
            #
            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["blur0"],div=2))
            self.interQuads[0].setShaderInput("src", self.getTextureInfo("color").texture)

            # Create a quad that outputs to texture "blur1". Connect its shader input named "src"
            # to the texture "blur0" (which is the output texture of the previous step).
            #
            # Now there is no div=..., so this quad outputs at its input resolution. The input "blur0"
            # is already quarter-resolution; hence the output "blur1" will be, too.
            #
            self.interQuads.append(self.pipeline.manager.renderQuadInto(colortex=self.textures["blur1"]))
            self.interQuads[1].setShaderInput("src", self.getTextureInfo("blur0").texture)

    def make(self):
        # Compile and assign shaders for internal stages.
        #
        # These particular shaders have no compile-time parameters, but if they did,
        # those would be applied here by generating the shader code in the appropriate
        # manner based on their current values (the values can be obtained by simply
        # reading the properties corresponding to the appropriate Filter-level
        # compile-time parameters).
        #
        self.interQuads[0].setShader(self.loadShader("filter-blurx.sha"))
        self.interQuads[1].setShader(self.loadShader("filter-blury.sha"))

    def synthesize(self):
        txblur1 = self.getTextureInfo("blur1")
        k_blur_amount = self.getMangledName("k_blur_amount")
        code = "pixcolor = lerp(tex2D(%(k_txblur1)s, %(texcoord_txblur1)s.xy), pixcolor, %(k_blur_amount)s.x);\n" % \
                   { "k_txblur1" : txblur1.varname,
                     "texcoord_txblur1" : txblur1.texcoord,
                     "k_blur_amount" : k_blur_amount }
        return ("blurSharpenFilter", code, "// Blur/sharpen blend pass\n")


    @property
    def amount(self):
        """Effect strength, float. 1 is off; values in [0,1) blur, values in (1,2] sharpen."""
        return self._amount
    @amount.setter
    def amount(self, value):
        self._amount = value
        if self.finalQuad is not None:
            # This input goes to the compositing shader, so we must mangle the name, but without the k_ prefix,
            # to get the unique name of the *shader input*.
            #
            self.finalQuad.setShaderInput( self.getMangledName("blur_amount"), value )

